<?php

namespace Drupal\ui_patterns\Element;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element;
use Drupal\ui_patterns\SourceInterface;
use Drupal\ui_patterns\SourcePluginBase;

/**
 * Provides a Component Prop form builder element.
 *
 * The value of the form element contains following keys:
 *  [
 *    'props' => [
 *      ['props_id' =>
 *       ['source_id' => 'id', 'value' => 'Source value ']
 *      ]
 *    ],
 *  ]
 * The same component value is returned in the form element.
 *
 * Usage example:
 *
 * @code
 * $form['slots'] = [
 *   '#type' => 'component_props_form',
 *   '#default_value' => [
 *     'props' => [],
 *   ],
 * ];
 * @endcode
 *
 * @FormElement("component_props_form")
 */
class ComponentPropsForm extends ComponentFormBase {

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = get_class($this);
    return [
      '#input' => TRUE,
      '#multiple' => FALSE,
      '#default_value' => NULL,
      '#source_contexts' => [],
      '#process' => [
        [$class, 'buildForm'],
      ],
      '#theme_wrappers' => ['container'],
    ];
  }

  /**
   * Build props forms.
   */
  public static function buildForm(array &$element, FormStateInterface $form_state): array {
    $component = static::getComponent($element);
    $contexts = $element['#source_contexts'] ?? [];
    $props = $component->metadata->schema['properties'];
    if (empty($props)) {
      return $element;
    }
    $configuration = $element['#default_value']['props'] ?? [];
    foreach ($props as $prop_id => $prop) {
      $element[$prop_id] = static::buildPropForm($element, $form_state, $prop_id, $prop, $configuration[$prop_id] ?? [], $contexts);
    }
    if (count(Element::children($element)) === 0) {
      hide($element);
    }
    return $element;
  }

  /**
   * Get sources for a prop type.
   */
  protected static function getSources(string $prop_id, array $definition, array $configuration, array $source_contexts, array $form_array_parents): array {
    $prop_type = $definition['ui_patterns']['type_definition'];
    $source_plugin_manager = \Drupal::service("plugin.manager.ui_patterns_source");
    $source_ids = array_keys($source_plugin_manager->getDefinitionsForPropType($prop_type->getPluginId(), $source_contexts));
    $source_ids = array_combine($source_ids, $source_ids);
    if (empty($source_ids)) {
      return [];
    }
    return $source_plugin_manager->createInstances($source_ids, SourcePluginBase::buildConfiguration($prop_id, $definition, $configuration, $source_contexts, $form_array_parents));
  }

  /**
   * Get selected source plugin.
   */
  protected static function getSelectedSource(array $configuration, array $sources): ?SourceInterface {
    if (isset($configuration['source_id']) && $sources[$configuration['source_id']]) {
      return $sources[$configuration['source_id']];
    }
    return NULL;
  }

  /**
   * Get default source plugin.
   */
  protected static function getDefaultSource(string $prop_id, array $definition, array $configuration, array $source_contexts, array $form_array_parents): ?SourceInterface {
    $source_plugin_manager = \Drupal::service("plugin.manager.ui_patterns_source");
    $prop_type = $definition['ui_patterns']['type_definition'];
    $source_id = $source_plugin_manager->getPropTypeDefault($prop_type->getPluginId(), $source_contexts);
    if (!$source_id) {
      return NULL;
    }
    $plugin_configuration = SourcePluginBase::buildConfiguration($prop_id, $definition, $configuration, $source_contexts, $form_array_parents);
    return $source_plugin_manager->createInstance($source_id, $plugin_configuration);
  }

  /**
   * Build single prop form.
   */
  protected static function buildPropForm(array $element, FormStateInterface $form_state, string $prop_id, array $definition, array $configuration, array $source_contexts): array {
    $form_array_parents = $element["#array_parents"];
    $sources = static::getSources($prop_id, $definition, $configuration, $source_contexts, $form_array_parents);
    $selected_source = static::getSelectedSource($configuration, $sources);
    if (!$selected_source) {
      $selected_source = static::getDefaultSource($prop_id, $definition, $configuration, $source_contexts, $form_array_parents);
    }
    if (!$selected_source) {
      return [];
    }

    $form_array_parents[] = $prop_id;
    $form_array_parents[] = "source";
    $wrapper_id = static::getElementId($element, 'ui-patterns-prop-item-' . $prop_id);
    $source_selector = static::buildSourceSelector($sources, $selected_source, $wrapper_id);
    $source_form = static::getSourcePluginForm($form_state, $selected_source, $wrapper_id, $form_array_parents);
    $prop_type = $definition['ui_patterns']['type_definition'];
    $build = [
      '#type' => 'fieldset',
      '#title' => $definition["title"] ?? $prop_type->label(),
      '#attributes' => [
        'style' => "position: relative;",
      ],
      'source_id' => $source_selector,
      'source' => $source_form,
    ];
    return $build;
  }

  /**
   * Get source plugin form.
   */
  protected static function getSourcePluginForm(FormStateInterface $form_state, SourceInterface $source, string $wrapper_id, array $form_array_parents): array {
    $form = $source->settingsForm([], $form_state);
    $form["#type"] = 'container';
    $form['#attributes'] = [
      'id' => $wrapper_id,
    ];
    // Weird, but :switchSourceForm() AJAX handler doesn't work without that.
    foreach (Element::children($form) as $child) {
      if (isset($form[$child]['#description']) && !isset($form[$child]['#description_display'])) {
        $form[$child]['#description_display'] = 'after';
      }
    }
    return $form;
  }

  /**
   * Build sources selector widget.
   */
  protected static function buildSourceSelector(array $sources, SourceInterface $selected_source, string $wrapper_id): array {
    if (empty($sources)) {
      return [];
    }
    if (count($sources) == 1) {
      return [
        '#type' => 'hidden',
        '#value' => array_keys($sources)[0],
      ];
    }
    $options = [];
    foreach ($sources as $source_id => $source) {
      $options[$source_id] = $source->label();
    }
    return [
      '#type' => 'select',
      "#options" => $options,
      '#default_value' => $selected_source->getPluginId(),
      '#attributes' => [
        'style' => "position: absolute; top: 0; right: 0;",
      ],
      '#prop_id' => $selected_source->getPropId(),
      '#prop_definition' => $selected_source->getPropDefinition(),
      '#ajax' => [
        'callback' => [
          static::class,
          'switchSourceForm',
        ],
        'wrapper' => $wrapper_id,
        'effect' => 'fade',
      ],
    ];
  }

  /**
   * Ajax handler: Switch source plugin form.
   */
  public static function switchSourceForm(array $form, FormStateInterface $form_state): array {
    $selector = $form_state->getTriggeringElement();
    $parents = $selector["#array_parents"];
    $subform = NestedArray::getValue($form, array_slice($parents, 0, -1));
    return $subform["source"];
  }

}
