<?php

declare(strict_types=1);

namespace Drupal\ui_patterns\Plugin\UiPatterns\Source;

use Drupal\Component\Render\MarkupInterface;
use Drupal\Component\Utility\Html;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Block\BlockPluginInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformState;
use Drupal\Core\Plugin\ContextAwarePluginInterface;
use Drupal\Core\Plugin\PluginWithFormsInterface;
use Drupal\ui_patterns\SourcePluginBase;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin implementation of the source.
 *
 * @Source(
 *   id = "block",
 *   label = @Translation("Block"),
 *   description = @Translation("instantiate a block"),
 *   prop_types = {
 *     "slot"
 *   }
 * )
 */
class BlockSource extends SourcePluginBase {
  /**
   * The block manager.
   *
   * @var \Drupal\Core\Block\BlockManagerInterface
   */
  protected $blockManager;


  /**
   * The Entity Type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;


  /**
   * The Entity Form builder.
   *
   * @var \Drupal\Core\Entity\EntityFormBuilderInterface
   */
  protected $entityFormBuilder;

  /**
   * The plugin form manager.
   *
   * @var \Drupal\Core\Plugin\PluginFormFactoryInterface
   */
  protected $pluginFormFactory;


  /**
   * The context handler.
   *
   * @var \Drupal\Core\Plugin\Context\ContextHandlerInterface
   */
  protected $contextHandler;

  /**
   * Block to be rendered.
   *
   * @var \Drupal\Core\Block\BlockPluginInterface|null
   */
  protected $block;

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition,
  ) {
    $plugin = parent::create(
      $container,
      $configuration,
      $plugin_id,
      $plugin_definition
    );
    $plugin->blockManager = $container->get('plugin.manager.block');
    $plugin->entityTypeManager = $container->get('entity_type.manager');
    $plugin->entityFormBuilder = $container->get('entity.form_builder');
    $plugin->pluginFormFactory = $container->get('plugin_form.factory');
    $plugin->contextHandler = $container->get('context.handler');
    $plugin->block = NULL;
    return $plugin;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultSettings(): array {
    return [
      'plugin_id' => NULL,
      'settings' => [],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getPropValue(): mixed {
    $plugin_id = $this->getSetting('plugin_id') ?? '';
    if (!$plugin_id) {
      return [];
    }
    $plugin_settings = $this->getSetting('settings') ?? [];
    // Create a block entity.
    $this->block = $this->getBlock($plugin_id, $plugin_settings, $this->contextRepository->getAvailableContexts());
    if (!$this->block) {
      return [];
    }
    return $this->block->build();
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    $plugin_id = $form_state->getValue('plugin_id');
    if ($plugin_id) {
      $plugin_settings = $this->getSetting('settings') ?? [];
      // Create a block entity.
      $block = $this->getBlock($plugin_id, $plugin_settings, $this->contextRepository->getAvailableContexts());
      $subform_state = SubformState::createForSubform($form['settings'], $form, $form_state);
      $this->getPluginForm($block)->validateConfigurationForm($form['settings'], $subform_state);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state): array {
    $form = parent::settingsForm($form, $form_state);
    $this->buildBlockCreateForm($form, $form_state);
    return $form;
  }

  /**
   * Build the form to create a block.
   */
  protected function buildBlockCreateForm(array &$form, FormStateInterface $form_state) {
    $definitions = $this->listBlockDefinitions();
    $options = $this->getBlockOptions($definitions);
    $wrapper_id = Html::getId(implode("_", $this->formArrayParents ?? []) . "_block-create-form-ajax");
    $plugin_id = $this->getSetting('plugin_id') ?? '';
    $form["plugin_id"] = [
      "#type" => "select",
      "#title" => t("Block"),
      "#options" => $options,
      '#default_value' => $plugin_id,

      '#ajax' => [
        'callback' => [__CLASS__, 'onBlockPluginIdChange'],
        'wrapper' => $wrapper_id,
        'method' => 'replace',
      // 'callback' => [static::class, 'onBlockPluginIdChange'],
      ],
      '#executes_submit_callback' => FALSE,
      '#empty_value' => '',
      '#empty_option' => t('- None -'),
      '#required' => TRUE,
    ];
    $form["settings"] = [
      '#type' => 'container',
      '#attributes' => ["id" => $wrapper_id],
      "#tree" => TRUE,
    ];
    // @todo support block settings form validation.
    // Ass seen from
    // \Drupal\layout_builder\Form\ConfigureBlockFormBase::validateForm
    if ($plugin_id) {

      $plugin_settings = $this->getSetting('settings') ?? [];
      // Create a block entity.
      $block = $this->getBlock($plugin_id, $plugin_settings, $this->contextRepository->getAvailableContexts());
      $form['#tree'] = TRUE;
      // $form['#process'] = [ '::validateForm'];
      $form['settings'] = [];
      $subform_state = SubformState::createForSubform($form['settings'], $form, $form_state);
      $form['settings'] = $this->getPluginForm($block)->buildConfigurationForm($form['settings'], $subform_state);

      /*
      $entity = $this->entityTypeManager->getStorage('block')->create(
      ['plugin' => $plugin_id, 'settings' => $plugin_settings]);
      $blockForm = $this->entityFormBuilder->getForm($entity);
      $blockSettingsForm = $blockForm["settings"];
      unset($blockSettingsForm["label"]);
      unset($blockSettingsForm["label_display"]);
      unset($blockSettingsForm["admin_label"]);
      $form["settings"] = $blockSettingsForm; */
      $form["settings"]['#tree'] = TRUE;
      $form['settings']['#prefix'] = '<div id="' . $wrapper_id . '">' . ($form['settings']['#prefix'] ?? '');
      $form['settings']['#suffix'] = ($form['settings']['#suffix'] ?? '') . '</div>';
    }
  }

  /**
   * Retrieves the plugin form for a given block.
   *
   * @param \Drupal\Core\Block\BlockPluginInterface $block
   *   The block plugin.
   *
   * @return \Drupal\Core\Plugin\PluginFormInterface
   *   The plugin form for the block.
   */
  protected function getPluginForm(BlockPluginInterface $block) {
    if ($block instanceof PluginWithFormsInterface) {
      return $this->pluginFormFactory->createInstance($block, 'configure');
    }
    return $block;
  }

  /**
   * Gets the plugin for this component.
   *
   * @param string $plugin_id
   *   The plugin ID.
   * @param array<mixed> $block_configuration
   *   The block configuration.
   * @param \Drupal\Core\Plugin\Context\ContextInterface[] $contexts
   *   An array of contexts to set on the plugin.
   *
   * @return \Drupal\Core\Block\BlockPluginInterface
   *   The plugin.
   */
  public function getBlock($plugin_id, $block_configuration, array $contexts = []) {
    $plugin = $this->blockManager->createInstance($plugin_id, $block_configuration);
    if ($contexts && $plugin instanceof ContextAwarePluginInterface) {
      $this->contextHandler->applyContextMapping($plugin, $contexts);
    }
    return $plugin;
  }

  /**
   * Ajax callback for fields with AJAX callback to update form substructure.
   *
   * @param array $form
   *   The form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   *
   * @return array
   *   The replaced form substructure.
   */
  public static function onBlockPluginIdChange(array $form, FormStateInterface $form_state): array {
    $triggeringElement = $form_state->getTriggeringElement();
    // Dynamically return the dependent ajax for elements based on the
    // triggering element. This shouldn't be done statically because
    // settings forms may be different, e.g. for layout builder, core, ...
    if (!empty($triggeringElement['#array_parents'])) {
      $subformKeys = $triggeringElement['#array_parents'];
      // Remove the triggering element itself and add the 'block' below key.
      array_pop($subformKeys);
      $subformKeys[] = 'settings';
      // Return the subform:
      $subform = NestedArray::getValue($form, $subformKeys);
      // $form_state->setRebuild();
      return $subform;
    }
    return [];
  }

  /**
   * Return blocks list.
   *
   * @see BlockLibraryController::listBlocks()
   * @see FilteredPluginManagerTrait::getFilteredDefinitions()
   * @see layout_builder_plugin_filter_block__block_ui_alter()
   * @see layout_builder_plugin_filter_block__layout_builder_alter()
   * @see layout_builder_plugin_filter_block_alter()
   *
   * @return array
   *   Definitions of blocks
   */
  protected function listBlockDefinitions() : array {
    $definitions = $this->blockManager->getFilteredDefinitions('ui_patterns', $this->contextRepository->getAvailableContexts(), []);
    // Order by category, and then by admin label.
    $definitions = $this->blockManager->getSortedDefinitions($definitions);
    // Filter out definitions that are not intended to be placed by the UI.
    $definitions = array_filter($definitions, function (array $definition) {
      return empty($definition['_block_ui_hidden']);
    });
    return $definitions;
  }

  /**
   * Get options for block select.
   */
  protected function getBlockOptions(array $definitions) : array {
    $definition_groups = [];
    foreach ($definitions as $plugin_id => $plugin_definition) {
      $category = $plugin_definition['category'] ?? 'Other';
      if ($category instanceof MarkupInterface) {
        /** @var \Drupal\Component\Render\MarkupInterface $category */
        $category = $category->__toString();
      }
      if (!array_key_exists($category, $definition_groups)) {
        $definition_groups[$category] = [];
      }
      $definition_groups[$category][$plugin_id] = $plugin_definition;
    }
    $options = [];
    foreach ($definition_groups as $definition_group_id => $definition_group) {
      $group_options = [];
      foreach ($definition_group as $definition_id => $definition) {
        $group_options[$definition_id] = $definition['admin_label'];
      }
      $options[$definition_group_id] = $group_options;
    }
    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function alterComponent(array $element): array {
    if (!$this->block) {
      return $element;
    }

    // CacheableMetadata::createFromRenderArray($content)
    $cache = $element["#cache"] ?? [];
    $element["#cache"] = array_merge($cache, [
      "max-age" => 0,
     // "tags" => ['config:system.menu.' . $this->menuId],
    ]);
    return $element;
  }

}
